unit ConvertData;

interface

uses Classes, TypInfo, Windows, SysUtils;

type
  TAfterConvertEvent=procedure(ASender:TObject) of object;
  TAfterConvertRecordEvent=procedure(ASender:TObject) of object;
  TAfterConvertFieldEvent=procedure(ASender:TObject; AFieldName:string; AFieldType:TTypeKind) of object;
  TConvertFieldEvent=procedure(ASender:TObject; AFieldName:string; AFieldType:TTypeKind; AFieldTypeData:PTypeData; var AValue:variant) of object;
  TBeforeConvertFieldEvent=procedure(ASender:TObject; AFieldName:string; AFieldType:TTypeKind) of object;
  TBeforeConvertRecordEvent=procedure(ASender:TObject) of object;
  TBeforeConvertEvent=procedure(ASender:TObject) of object;

  TDataReader=class
  private
    FStream: TStream;
    function GetReadable: boolean;
  public
    property Stream:TStream read FStream;
    property Readable:boolean read GetReadable;
    procedure readField(AFieldName:string; AFieldType:TTypeKind; AFieldTypeData:PTypeData; out AFieldValue:variant); virtual; abstract;
    constructor Create(AStream:TStream);
  end;

  TDataWriter=class
  private
    FStream: TStream;
  public
    property Stream:TStream read FStream;
    procedure writeField(AFieldName:string; AFieldType:TTypeKind; AFieldTypeData:PTypeData; AFieldValue:variant); virtual; abstract;
    constructor Create(AStream:TStream);
  end;

  {$M+}
  TBaseDataConverter=class
  private
    FOnAfterConvert: TAfterConvertEvent;
    FOnAfterConvertField: TAfterConvertFieldEvent;
    FOnBeforeConvert: TBeforeConvertEvent;
    FOnBeforeConvertField: TBeforeConvertFieldEvent;
    FOnAfterConvertRecord: TAfterConvertRecordEvent;
    FOnBeforeConvertRecord: TBeforeConvertRecordEvent;
    FOnConvertField: TConvertFieldEvent;
    procedure DoBeforeConvert; virtual;
    procedure DoBeforeConvertRecord; virtual;
    procedure DoBeforeConvertField(AFieldName:string; AFieldType:TTypeKind); virtual;
    procedure DoConvertField(AFieldName:string; AFieldType:TTypeKind; AFieldTypeData:PTypeData; var AValue:variant); virtual;
    procedure DoAfterConvertField(AFieldName:string; AFieldType:TTypeKind); virtual;
    procedure DoAfterConvertRecord; virtual;
    procedure DoAfterConvert; virtual;
  protected
    FSource:TDataReader;
    FDestination:TDataWriter;
  public
    constructor Create(ASource:TDataReader; ADestination:TDataWriter);
    procedure Convert;

    property OnBeforeConvert:TBeforeConvertEvent read FOnBeforeConvert write FOnBeforeConvert;
    property OnAfterConvert:TAfterConvertEvent read FOnAfterConvert write FOnAfterConvert;
    property OnBeforeConvertRecord:TBeforeConvertRecordEvent read FOnBeforeConvertRecord write FOnBeforeConvertRecord;
    property OnAfterConvertRecord:TAfterConvertRecordEvent read FOnAfterConvertRecord write FOnAfterConvertRecord;
    property OnBeforeConvertField:TBeforeConvertFieldEvent read FOnBeforeConvertField write FOnBeforeConvertField;
    property OnAfterConvertField:TAfterConvertFieldEvent read FOnAfterConvertField write FOnAfterConvertField;
    property OnConvertField:TConvertFieldEvent read FOnConvertField write FOnConvertField;
  end;
  {$M-}

  TBinaryReader=class(TDataReader)
  public
    procedure readField(AFieldName: String; AFieldType: TTypeKind; AFieldTypeData:PTypeData; out AFieldValue: Variant); override;
  end;

  TCSVWriter=class(TDataWriter)
  public
    procedure writeField(AFieldName: String; AFieldType: TTypeKind; AFieldTypeData:PTypeData; AFieldValue: Variant); override;
  end;

implementation

uses Variants;

{ TBaseDataConverter }

constructor TBaseDataConverter.Create(ASource:TDataReader; ADestination:TDataWriter);
begin
  FSource:=ASource;
  FDestination:=ADestination;
end;

procedure TBaseDataConverter.DoAfterConvert;
begin
  if assigned(FOnAfterConvert) then
    FOnAfterConvert(self);
end;

procedure TBaseDataConverter.DoAfterConvertField(AFieldName: string;
  AFieldType: TTypeKind);
begin
  if assigned(FOnAfterConvertField) then
    FOnAfterConvertField(self, AFieldName, AFieldType);
end;

procedure TBaseDataConverter.DoAfterConvertRecord;
begin
  if assigned(FOnAfterConvertRecord) then
    FOnAfterConvertRecord(self);
end;

procedure TBaseDataConverter.DoBeforeConvert;
begin
  if assigned(FOnBeforeConvert) then
    FOnBeforeConvert(self);
end;

procedure TBaseDataConverter.DoBeforeConvertField(AFieldName: string;
  AFieldType: TTypeKind);
begin
  if assigned(FOnBeforeConvertField) then
    FOnBeforeConvertField(self, AFieldName, AFieldType);
end;

procedure TBaseDataConverter.DoBeforeConvertRecord;
begin
  if assigned(FOnBeforeConvertRecord) then
    FOnBeforeConvertRecord(self);
end;

procedure TBaseDataConverter.DoConvertField(AFieldName: string;
  AFieldType: TTypeKind; AFieldTypeData:PTypeData; var AValue: variant);
begin
  if assigned(FOnConvertField) then
    FOnConvertField(Self, AFieldName, AFieldType, AFieldTypeData, AValue);
end;

procedure TBaseDataConverter.Convert;
var i, c:integer;
    pis:PPropList;
    pi:PPropInfo;
    td:PTypeData;
    v:variant;
begin
  c:=GetPropList(ClassInfo, tkProperties, nil, false);
  GetMem(pis, c * SizeOf(PPropInfo));
  try
    GetPropList(ClassInfo, tkProperties, pis, false);
    DoBeforeConvert;
    while FSource.Readable do
    begin
      DoBeforeConvertRecord;
      for i:=1 to c do
      begin
        pi:=pis^[i-1];
        assert(assigned(pi));
        td:=GetTypeData(pi^.PropType^);
        DoBeforeConvertField(pi.Name, pi.PropType^.Kind);
        FSource.readField(pi.Name, pi.PropType^.Kind, td, v);
        DoConvertField(pi.Name, pi.PropType^.Kind, td, v);
        FDestination.writeField(pi.Name, pi.PropType^.Kind, td, v);
        DoAfterConvertField(pi.Name, pi.PropType^.Kind);
      end;
      DoAfterConvertRecord;
    end;
    DoAfterConvert;
  finally
    FreeMem(pis);
  end;
end;

{ TDataWriter }

constructor TDataWriter.Create(AStream: TStream);
begin
  FStream:=AStream;
end;

{ TDataReader }

constructor TDataReader.Create(AStream: TStream);
begin
  FStream:=AStream;
end;

function TDataReader.GetReadable: boolean;
begin
  result:=FStream.Position<FStream.Size;
end;

{ TBinaryReader }

procedure TBinaryReader.readField(AFieldName: String;
  AFieldType: TTypeKind; AFieldTypeData:PTypeData; out AFieldValue: Variant);
var sb:Shortint;
    ub:byte;
    sw:Smallint;
    uw:word;
    sl:Longint;
    ul:LongWord;
    s:string;
begin
  case AFieldType of
    tkInteger:
      begin
        case AFieldTypeData.OrdType of
          otSByte:
            begin
              FStream.Read(sb, 1);
              AFieldValue:=sb;
            end;
          otUByte:
            begin
              FStream.Read(ub, 1);
              AFieldValue:=ub;
            end;
          otSWord:
            begin
              FStream.Read(sw, 2);
              AFieldValue:=sw;
            end;
          otUWord:
            begin
              FStream.Read(uw, 2);
              AFieldValue:=uw;
            end;
          otSLong:
            begin
              FStream.Read(sl, 4);
              AFieldValue:=sl;
            end;
          otULong:
            begin
              FStream.Read(ul, 4);
              AFieldValue:=ul;
            end;
        end;
      end;
    tkString:
      begin
        assert(AFieldTypeData.MaxLength>0);// record strings are fixed length and have a  maximum of 255 chars
        setlength(s, AFieldTypeData.MaxLength);
        FStream.read(pchar(s)^, AFieldTypeData.MaxLength);
        AFieldValue:=s;
      end;
    else
      raise Exception.CreateFmt('Field type %s not implemented by Binary Reader', [GetEnumName(TypeInfo(TTypeKind), integer(AFieldType))]);
  end;
end;

{ TCSVWriter }

procedure TCSVWriter.writeField(AFieldName: String; AFieldType: TTypeKind;
  AFieldTypeData:PTypeData; AFieldValue: Variant);
var s:string;
begin
  if not VarIsStr(AFieldValue) then
    raise Exception.Create('CSV writer expects already formatted string data.');
  s:=vartostr(AFieldValue);
  FStream.write(pchar(s)^, length(s));
end;

end.
